(define-module (ui mastercontrol dispatch ajax)
  :use-module (alterator ajax)
  :use-module (alterator ahttpd)
  :use-module (alterator algo)
  :use-module (alterator plist)
  :use-module (alterator woo)
  :export (init))

(define *hostlist* '())
(define *total* 0)
(define *processed* 0)
(define *url-ok* #f)
(define *branch-ok* #f)
(define *key-ok* #f)

(define (totalof total processed)
  (format #f (_ "~a of ~a") (- total processed) total))

(define (progress total processed)
  (round (* 100 (if (> total 0) (/ processed total) 0))))

(define (update-counters total . args)
  (let ((total total)
        (processed (if (null? args) 0 (car args)))
        (success (if (< (length args) 2) 0 (cadr args)))
        (failed (if (< (length args) 3) 0 (caddr args)))
        (unknown (if (< (length args) 4) 0 (cadddr args))))
    (set! *total* total)
    (set! *processed* processed)
    (list total processed success failed unknown)))

(define (update-counters-form total processed success failed unknown)
  (form-update-value "total" total)
  (form-update-value "totalof" (totalof total processed))
  (form-update-value "progress" (progress total processed))
  (form-update-value "successful" success)
  (form-update-value "failed" failed)
  (form-update-value "unknown" unknown))

(define (read-stats)
  (catch/message
    (lambda ()
      (let ((data (woo-read-first (string-append "/mastercontrol/hosts/stats/" (ahttpd-session-ref 'mastercontrol-source))
                                  'host (ahttpd-session-ref 'mastercontrol-host)
                                  'hoststatus (ahttpd-session-ref 'mastercontrol-hoststatus)
                                  'group (ahttpd-session-ref 'mastercontrol-hostgroup)
                                  'filtersql (ahttpd-session-ref 'mastercontrol-filtersql))))
        (apply update-counters-form (update-counters (string->number (woo-get-option data 'total))))))))

(define (update-hostlist hostlist)
  (set! *hostlist* (string-split hostlist #\;))
  (apply update-counters-form (update-counters (length *hostlist*))))

(define (update-branch branch)
  (form-update-value "profile" branch)
  (form-update-visibility "select-profile" branch)
  (form-update-visibility "select-sshkey" branch)
  (form-update-value "fetch-result" (if branch
                                      (_ "Successfully fetched the profile list")
                                      (_ "No profiles are published at the given URL")))
  (set! *branch-ok* branch))

(define (update-delete-key)
  (catch/message
    (lambda ()
      (let ((data (woo-read-first "/mastercontrol/sshkey/dir" 'sshkey (form-value "sshkey"))))
        (form-update-activity "delete-key" (equal? (woo-get-option data 'dir) "mastercontrol"))))))

(define (update-dispatch)
  (form-update-activity "dispatch" (and *url-ok* *branch-ok* *key-ok*)))

(define (update-buttons)
  (update-delete-key)
  (update-dispatch))

(define (read-keys)
  (catch/message
    (lambda ()
      (set! *key-ok* #f)
      (form-update-activity "delete-key" #f)
      (let* ((keys (woo-list "/mastercontrol/sshkeys/private"))
             (key (and (not (null? keys))
                       (cdr (plistq 'name (cdar keys))))))
        (form-update-enum "sshkey" keys)
        (form-update-value "sshkey" key)
        (set! *key-ok* key))))
  (update-buttons)
  *key-ok*)

(define (fetch url)
  (catch/message
    (lambda ()
      (update-branch #f)
      (set! *url-ok* #f)
      (let* ((branches (woo-list "/mastercontrol/branches" 'url url))
             (branch (and (not (null? branches))
                          (cdr (plistq 'name (cdar branches))))))
        (set! *url-ok* #t)
        (form-update-enum "profile" branches)
        (update-branch branch))
      (and *url-ok* *branch-ok* (read-keys)))))

(define (upload-key)
  (let* ((res (js-result 'askUpload (_ "SSH key upload") (_ "Upload") (_ "Cancel")))
         (sshkey-file (cond-assoc 'sshkey-file res)))
    (if (and sshkey-file (not (string-null? sshkey-file)))
      (begin
        (ahttpd-session-set! 'ahttpd-full-args (cons (cons "sshkey-file" sshkey-file) (ahttpd-session-ref 'ahttpd-full-args)))
        (call-with-form-file "sshkey-file"
          (lambda (path)
            (if (and path (not (string-null? path)))
              (let* ((res (js-result 'askPassword (_ "SSH private key password") (_ "OK") (_ "Cancel"))))
                (catch/message
                  (lambda ()
                    (woo-write "/mastercontrol/sshkey/private" 'sshkey path 'passwd (cond-assoc 'sshpasswd res))))))))
        (ahttpd-session-set! 'ahttpd-full-args (cons (cons "sshkey-file" #f) (ahttpd-session-ref 'ahttpd-full-args)))
        (read-keys)))))

(define (delete-key)
  (if (form-confirm (_ "Are you sure to delete the selected private key?")
                    (_ "Delete SSH key"))
    (catch/message
      (lambda ()
        (woo-write "/mastercontrol/sshkey/private" 'sshkey (form-value "sshkey") 'delete #t)
        (read-keys)))))

(define (read-counters)
  (catch/message
    (lambda ()
      (let* ((data (woo-read-first "/mastercontrol/dispatch"))
             (total (string->number (woo-get-option data 'total)))
             (ok (string->number  (woo-get-option data 'ok)))
             (modified (string->number  (woo-get-option data 'modified)))
             (unknown (string->number  (woo-get-option data 'unknown)))
             (disabled (string->number  (woo-get-option data 'disabled)))
             (other (string->number  (woo-get-option data 'other)))
             (failed (string->number  (woo-get-option data 'failed)))
             (processed (string->number  (woo-get-option data 'processed)))
             (success (if (> ok (- processed failed)) (- processed failed) ok))
             (unsuccess (if (> (- processed failed ok)
                               (+ modified unknown disabled other))
                          (+ modified unknown disabled other)
                          (if (> (- processed failed ok) 0)
                            (- processed failed ok)
                            0))))
        (list total processed success failed unsuccess)))))

(define (read-counters-loop)
  (apply (lambda (total processed success failed unknown)
           (apply update-counters-form (update-counters total processed success failed unknown))
           (if (< processed total)
             (form-set-timeout read-counters-loop 2)
             (begin
               (form-update-activity "progress-ok" #t)
               (form-update-activity "progress-cancel" #f)
               (cond
                 ((> unknown 0)
                  (form-update-activity "progress-refresh" #t)
                  (form-update-value "dispatch-result" (_ "Dispatch finished. Use the \"Refresh\" button to update the statistics above")))
                 ((and (> success 0) (> failed 0))
                  (form-update-value "dispatch-result" (_ "Dispatch finished. Some updates failed")))
                 ((and (= success total))
                  (form-update-value "dispatch-result" (_ "Dispatch successful")))
                 (else
                   (form-update-value "dispatch-result" (_ "Dispatch failed")))))))
         (read-counters)))

(define (dispatch)
  (form-update-activity "progress-ok" #f)
  (form-update-activity "progress-cancel" #t)
  (form-update-activity "progress-refresh" #f)
  (form-update-value "dispatch-result" "")
  (form-update-activity "fetch" #f)
  (form-update-activity "upload-key" #f)
  (form-update-activity "delete-key" #f)
  (form-update-visibility "progressbar" #t)
  (form-update-activity "dispatch" #f)
  (let* ((res (js-result 'askPassword (_ "SSH private key password") (_ "OK") (_ "Cancel"))))
   (and res (cond-assoc 'retcode res)
    (catch/message
      (lambda ()
        (if (null? *hostlist*)
          (woo-write (string-append "/mastercontrol/dispatch/" (ahttpd-session-ref 'mastercontrol-source))
                     'host (ahttpd-session-ref 'mastercontrol-host)
                     'hoststatus (ahttpd-session-ref 'mastercontrol-hoststatus)
                     'group (ahttpd-session-ref 'mastercontrol-hostgroup)
                     'filtersql (ahttpd-session-ref 'mastercontrol-filtersql)
                     'sshkey (form-value "sshkey")
                     'passwd (cond-assoc 'sshpasswd res)
                     'url (form-value "url")
                     'head (form-value "profile"))
          (woo-write "/mastercontrol/dispatch"
                     'hostlist (string-join *hostlist* " ")
                     'sshkey (form-value "sshkey")
                     'passwd (cond-assoc 'sshpasswd res)
                     'url (form-value "url")
                     'head (form-value "profile")))
        (read-counters-loop))))))

(define (stop-dispatch)
  (catch/message
    (lambda ()
      (woo-write "/mastercontrol/dispatch" 'head #f))))

(define (init)
  (form-bind "fetch" "click"
    (lambda ()
      (fetch (form-value "url"))))
  (form-bind "url" "enter"
    (lambda ()
      (fetch (form-value "url"))))
  (form-bind "sshkey" "change" update-buttons)
  (form-bind "upload-key" "click" upload-key)
  (form-bind "delete-key" "click" delete-key)
  (form-bind "dispatch" "click" dispatch)
  (form-bind "progress-ok" "click"
    (lambda ()
      (stop-dispatch)
      (form-replace "/mastercontrol")))
  (form-bind "progress-cancel" "click"
    (lambda ()
      (stop-dispatch)
      (form-replace "/mastercontrol" 'hostlist (form-value "hostlist"))))
  (form-bind "back" "click"
    (lambda ()
      (stop-dispatch)
      (form-replace "/mastercontrol" 'hostlist (form-value "hostlist"))))
  (form-bind "progress-refresh" "click"
    (lambda ()
      (apply update-counters-form (apply update-counters (read-counters)))))
  (if (equal? (form-value "update-range") "all")
    (begin
      (set! *hostlist* '())
      (read-stats))
    (begin
      (update-hostlist (form-value "hostlist"))))
  (form-update-visibility "select-profile" #f)
  (form-update-visibility "select-sshkey" #f)
  (form-update-activity "dispatch" #f))
